

"""
-> if CurrentCharacter is '.'
    - > if three previous characters are not periods
    - > if first next character is a space
    - > if second next character is a sentence beginner
"""
# This one is actually an extremely difficult issue to solve.
from FNLP.Language import Character
from F import MATH
from FNLP.Language import Constants

FORM_SENTENCE = lambda strContent, startIndex, endIndex, caboose: f"{strContent[startIndex:endIndex]}{caboose}"

def to_sentences(content: str, combineQuotes=True):
    """100%! WORKING!!!"""
    content = __prepare_content_for_sentence_extraction(content)
    current_index, start_index, quotation_count, sentences = 0, 0, 0, []
    # -> Loop every character in content (str).
    for currentChar in content:
        # -> Verify we have next (+3) characters.
        if current_index >= len(content) - 3:
            if MATH.is_even_number(quotation_count + 1):
                sent = content[start_index:-1] + currentChar + '"'
            else:
                sent = content[start_index:-1] + currentChar
            sentences.append(sent)
            break
        plusOneChar = content[current_index + 1]
        # -> Keep count of quotations. Even = Closed and Odd = Open
        if Character.is_quotation(currentChar):
            quotation_count += 1
        # -> Verify current (0) character is a "sentence ending character" EX: . ! ?
        if __is_sentence_ender(currentChar):
            # -> Verify next (+1) character is a space or quote "
            if Character.is_space(plusOneChar) or Character.is_quotation(plusOneChar):
                QM = False
                # -> Verify next (+1) character is a quotation and are we "Quote Closed"
                if Character.is_quotation(plusOneChar) and MATH.is_even_number(quotation_count + 1):
                    QM = True
                plusTwoChar = str(content[current_index + 2])
                plusThreeChar = str(content[current_index + 3])
                # -> Verify that +2 character is a valid "Sentence Beginner"
                if __is_sentence_beginner(plusTwoChar if not QM else plusThreeChar):
                    # -> Verify next character is a space
                    if Character.is_space(plusOneChar if not QM else plusTwoChar):
                        minusOneChar = str(content[current_index - 1])
                        minusTwoChar = str(content[current_index - 2])
                        minuxThreeChar = str(content[current_index - 3])
                        # -> Verify previous three characters do not include a period
                        if not Character.are_periods(minusOneChar, minusTwoChar, minuxThreeChar):
                            if combineQuotes and not MATH.is_even_number(quotation_count if not QM else quotation_count + 1):
                                current_index += 1
                                continue
                            # -> VERIFIED! Create sentence.
                            sent = FORM_SENTENCE(content, start_index, current_index, currentChar if not QM else f'{currentChar}"')
                            start_index = current_index + 2
                            sentences.append(sent)
        current_index += 1
    return sentences


def __prepare_content_for_sentence_extraction(content):
    return content.strip().replace("\n", " ").replace("  ", " ")

def __is_sentence_ender(content):
    if str(content) in Constants.SENTENCE_ENDERS:
        return True
    return False

def __is_sentence_beginner(content):
    if Character.is_in_alphabet(content):
        return True
    elif Character.is_quotation(content):
        return True
    elif Character.is_single_number(content):
        return True
    return False


from . import Words
from . import Constants

def keywords(content):
    """Get the top 10 keywords and their frequency scores ignores blacklisted
    words in stopwords, counts the number of occurrences of each word, and
    sorts them in reverse natural order (so descending) by number of
    occurrences.
    """
    NUM_KEYWORDS = 10
    content = Words.__split_words(content)
    # of words before removing blacklist words
    if content:
        num_words = len(content)
        content = [x for x in content if x not in Constants.STOP_WORDS]
        freq = {}
        for word in content:
            if word in freq:
                freq[word] += 1
            else:
                freq[word] = 1

        min_size = min(NUM_KEYWORDS, len(freq))
        keywords = sorted(freq.items(),
                          key=lambda x: (x[1], x[0]),
                          reverse=True)
        keywords = keywords[:min_size]
        keywords = dict((x, y) for x, y in keywords)
        for k in keywords:
            articleScore = keywords[k] * 1.0 / max(num_words, 1)
            keywords[k] = articleScore * 1.5 + 1
        return dict(keywords)
    else:
        return dict()





SEARCH_TERM = lambda searchTerm: fr'.*{str(searchTerm)}.*'
SEARCH_TERM_TWO = lambda searchTerm: fr'\b{searchTerm}\b'
REMOVE_SPECIAL_CHARACTERS = r'[^a-zA-Z0-9]'

CONTAINS_YEAR = r'\b[1-2][0-9][0-9][0-9]\b'
ONLY_CAPITAL_WORDS = r'([A-Z][a-z]+)'

EXTRACT_TICKERS_STRICT = r'[^\s$:(][A-Z]{1,5}(?=\s|\)|/|,|\.)'
EXTRACT_TICKERS_LIGHT = r'[^\s$:(][A-Za-z]{1,5}(?=\s|\)|/|,|\.)'

IS_URL = r'http.?://.*/'
FIND_URLS = 'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'



import F

from FNLP.Language import Words


class Sentence:
    id = F.get_uuid()
    rawText = ""
    orderNumber = None
    words = []

    def __init__(self, text, orderNumber=None):
        self.rawText = text
        self.orderNumber = orderNumber
        self.words = []

    def tokenize(self):
        self.words = Words.to_words_v1(self.rawText)
        return self.words

class Sentences:
    sentences: [Sentence]


"""

 0: { 
     0: "Sentence()",
     1: "sentence2",
     2: "sentence3",
     3: "sentence4",
     4: "sentence5"
    }
    
    
"""


import re
from FNLP.Language.Constants import STOP_WORDS
from FNLP.Language import Character


def extract_only_capital_words_regex(content):
    capital_words = re.findall(r'([A-Z][a-z]+)', content)
    return capital_words

def extract_only_capital_words_manual(content):
    previous_char = " "
    current_index = 0
    start_index = 0
    start_enabled = False
    temp_words = []
    for char in content:
        if Character.is_capital(char) and Character.is_empty(char) and not start_enabled:
            start_index = current_index
            start_enabled = True
        if not Character.is_capital(char) and Character.is_empty(char) and start_enabled:
            end_index = current_index
            words = content[start_index:end_index-1]
            temp_words.append(words)
            start_enabled = False
        previous_char = char
        current_index += 1
    final_words = []
    for word in temp_words:
        if word.lower() in STOP_WORDS:
            continue
        final_words.append(word)
    return final_words


from F import LIST

def remove_empty_strings(list_of_strs: []):
    newS = []
    for word in list_of_strs:
        if word == '':
            continue
        newS.append(word)
    return newS

def replace(content, *args):
    for arg in args:
        content = content.replace(arg, " ")
    return content

def combine_args_str(*content: str) -> str:
    temp = ""
    content = LIST.flatten(content)
    for item in content:
        temp += " " + str(item)
    return str(temp).strip()


from F import LIST, DICT
from FNLP.Language import Utils, Character, Constants
from FNLP.Regex import ReWords



""" 
- input: Takes in two lists to match
- output: ( Overall_Score, { matched_term: count } 
"""
def matcher(word_list: list, weighted_terms: list, score_variant=0.6) -> (int,{}):
    # 2. -> Loop Each Category Weighted Term
    temp_dict = {}  # { "weighted_term": "match_count" }
    score = 0  # "weighted_term" Score * "match_count"
    for w_term in weighted_terms:
        # Stay Safe People
        if not w_term or w_term == "" or w_term == " ":
            continue
        # -> Expand Weighted Term
        # expanded_key_list = expand_word(w_term)
        # 3. -> Loop All Tokens AND MATCH!!
        for token in word_list:
            # Stay Safe People
            if not token or token == "" or token == " ":
                continue
            # MATCHER! -> if content word is in expanded weighted term list...
            no_the_token = remove_the(token)
            score_phrase = phrase_match_percentage(token, w_term)
            score_no_the = phrase_match_percentage(no_the_token, w_term)
            if score_phrase >= score_variant or score_no_the >= score_variant:
            # if token == w_term or no_the_token == w_term:
                # -> We have a match!
                key_score = 1
                score += key_score
                temp_dict = DICT.add_matched_word_to_result(token, temp_dict)
    # -> 4. Finish Up
    return score, temp_dict  # ( score, { "weighted_term": "match_count", "weighted_term": "match_count" } )

def matcher_with_expander(word_list: list, weighted_terms: list) -> (int,{}):
    # 2. -> Loop Each Category Weighted Term
    temp_dict = {}  # { "weighted_term": "match_count" }
    score = 0  # "weighted_term" Score * "match_count"
    for w_term in weighted_terms:
        # Stay Safe People
        if not w_term or w_term == "" or w_term == " ":
            continue
        # -> Expand Weighted Term
        expanded_key_list = expand_word(w_term)
        # 3. -> Loop All Tokens AND MATCH!!
        for token in word_list:
            # Stay Safe People
            if not token or token == "" or token == " ":
                continue
            # MATCHER! -> if content word is in expanded weighted term list...
            if is_match(token, expanded_key_list):
                # -> We have a match!
                key_score = 1
                score += key_score
                temp_dict = DICT.add_matched_word_to_result(w_term, temp_dict)
    # -> 4. Finish Up
    return score, temp_dict  # ( score, { "weighted_term": "match_count", "weighted_term": "match_count" } )

def phrase_match_percentage(phrase_one, phrase_two):
    pListOne = to_words_v1(phrase_one)
    pListTwo = to_words_v1(phrase_two)
    poneCount = len(pListOne)
    ptwoCount = len(pListTwo)
    if poneCount > ptwoCount:
        highest = poneCount
    else:
        highest = ptwoCount
    result = 0
    for i in range(highest):
        poneword = LIST.get(i, pListOne, False)
        ptwoword = LIST.get(i, pListTwo, False)
        if poneword == ptwoword:
            result += 1
    r = result / highest
    return r

def find_proper_nouns(content:str):
    raw = ReWords.extract_only_capital_words_regex(content)
    proper_nouns = []
    for word in raw:
        if str(word).lower() in Constants.STOP_WORDS:
            continue
        proper_nouns.append(word)
    return proper_nouns

def find_proper_nouns_v2(content:str):
    tokens = to_words_v1(content)
    proper_nouns = []
    temp = []
    for word in tokens:
        if is_capital(word):
            if str(word).lower() in Constants.STOP_WORDS:
                continue
            temp.append(word)
        else:
            if temp:
                proper_nouns.append(temp)
                temp = []
    return proper_nouns

def is_match(word:str, word_list:list, ignoreCapitals=False):
    if ignoreCapitals:
        word = str(word).lower() # Make LowerCase
        word_list = [str(item).lower() for item in word_list] # Make All Lowercase
    if word in word_list:
        return True
    return False

def is_capital(word:str):
    firstChar = word[0]
    if str(firstChar).isupper():
        return True
    return False

def make_capital(word:str):
    firstChar = word[0]
    newWord = word[1:]
    if Character.is_in_alphabet(firstChar) and not Character.is_capital(firstChar):
        firstChar = Constants.GET_CAPITAL_FROM_LOWER(firstChar)
    newWord = firstChar + newWord
    return newWord

def make_lower(word:str):
    firstChar = word[0]
    newWord = word[1:]
    if Character.is_in_alphabet(firstChar) and Character.is_capital(firstChar):
        firstChar = Constants.GET_LOWER_FROM_CAPITAL(firstChar)
    newWord = firstChar + newWord
    return newWord

def to_x_grams(tokens, x):
    """ PUBLIC """
    if type(tokens) == str:
        tokens = to_words_v1(tokens)
    i = 0
    x_grams = []
    if len(tokens) < x:
        print("found none", tokens)
        return x_grams
    for _ in tokens:
        if i+x > len(tokens):
            break
        phrase = ""
        for c in range(x):
            phrase = combine_words(phrase, tokens[i+c])
        x_grams.append(phrase)
        i += 1
    return x_grams

def to_words_v1(content: str):
    content = Utils.replace(content, ".", ",", ";", "\n", "  ")
    s = content.split(" ")
    newS = Utils.remove_empty_strings(s)
    return newS

def to_bi_grams_v2(tokens):
    """ PUBLIC HELPER """
    return to_x_grams(tokens, 2)

def to_tri_grams_v2(tokens):
    """ PUBLIC HELPER """
    return to_x_grams(tokens, 3)

def to_quad_grams_v2(tokens):
    """ PUBLIC HELPER """
    return to_x_grams(tokens, 4)

def expand_word(word: str) -> []:
    """ PUBLIC -> FAIR Expansion <- """
    word_lower = word.lower()
    word_upper = word.upper()
    word_first_capital = word[0].upper() + word[1:]
    # word_stem = lemmatize_word(word)
    # word_stem_first_capital = word_stem[0].upper() + word_stem[1:]
    return [word, word_lower, word_upper, word_first_capital]

def score_words(words):
    result = {}
    for word in words:
        if word in result.keys():
            tempValue = result[word]
            result[word] = tempValue + 1
        else:
            result[word] = 1
    return result

def combine_words(*words):
    """ Combines two strings together. """
    temp_word = ""
    words = LIST.flatten(words)
    if len(words) > 0:
        for word in words:
            temp_word += " " + word.strip()
        return temp_word.strip()
    return str(words).strip()

def remove_ing(word):
    if word.endswith("ing"):
        return word[:-3]
    elif word.endswith("ings"):
        return word[:-4]
    return False

def remove_apos(word):
    word = word.replace("'", "")
    return word

def remove_the(word:str):
    word = word.strip()
    if word.startswith("the") or word.startswith("The"):
        word = word.replace("the", "").replace("The", "")
        return word.strip()
    return word

def __split_words(text):
    """ ALTERNATIVE: Split a string into array of words. """
    try:
        import re
        text = re.sub(r'[^\w ]', '', text)  # strip special chars
        return [x.strip('.').lower() for x in text.split()]
    except TypeError:
        return None


if __name__ == '__main__':
    wordOne = "Sandbox development"
    wordTwo = "The Sandbox"
    print(phrase_match_percentage(wordOne, wordTwo))